/*****************************************************************************/
/*                                                                           */
/*                                 output.h                                  */
/*                                                                           */
/*                           Output file handling                            */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/* (C) 2009-2012, Ullrich von Bassewitz                                      */
/*                Roemerstrasse 52                                           */
/*                D-70794 Filderstadt                                        */
/* EMail:         uz@cc65.org                                                */
/*                                                                           */
/*                                                                           */
/* This software is provided 'as-is', without any expressed or implied       */
/* warranty.  In no event will the authors be held liable for any damages    */
/* arising from the use of this software.                                    */
/*                                                                           */
/* Permission is granted to anyone to use this software for any purpose,     */
/* including commercial applications, and to alter it and redistribute it    */
/* freely, subject to the following restrictions:                            */
/*                                                                           */
/* 1. The origin of this software must not be misrepresented; you must not   */
/*    claim that you wrote the original software. If you use this software   */
/*    in a product, an acknowledgment in the product documentation would be  */
/*    appreciated but is not required.                                       */
/* 2. Altered source versions must be plainly marked as such, and must not   */
/*    be misrepresented as being the original software.                      */
/* 3. This notice may not be removed or altered from any source              */
/*    distribution.                                                          */
/*                                                                           */
/*****************************************************************************/



#ifndef OUTPUT_H
#define OUTPUT_H



#include <stdio.h>

/* common */
#include "attrib.h"



/*****************************************************************************/
/*                                   Data                                    */
/*****************************************************************************/



/* Name of the output file. Dynamically allocated and read only. */
extern const char* OutputFilename;

/* Output file handle. Use WriteOutput if possible. Read only. */
extern FILE* OutputFile;



/*****************************************************************************/
/*                                   Code                                    */
/*****************************************************************************/



void SetOutputName (const char* Name);
/* Sets the name of the output file. */

void MakeDefaultOutputName (const char* InputFilename);
/* If the name of the output file is empty or NULL, the name of the output
** file is derived from the input file by adjusting the file name extension.
*/

void OpenOutputFile ();
/* Open the output file. Will call Fatal() in case of failures. */

void OpenDebugOutputFile (const char* Name);
/* Open an output file for debugging purposes. Will call Fatal() in case of
** failures.
*/

void CloseOutputFile ();
/* Close the output file. Will call Fatal() in case of failures. */

int WriteOutput (const char* Format, ...) attribute ((format (printf, 1, 2)));
/* Write to the output file using printf like formatting. Returns the number
** of chars written.
*/


/* End of output.h */

#endif
