/*****************************************************************************/
/*                                                                           */
/*                                  color.h                                  */
/*                                                                           */
/*          Color definition for the sp65 sprite and bitmap utility          */
/*                                                                           */
/*                                                                           */
/*                                                                           */
/* (C) 2012,      Ullrich von Bassewitz                                      */
/*                Roemerstrasse 52                                           */
/*                D-70794 Filderstadt                                        */
/* EMail:         uz@cc65.org                                                */
/*                                                                           */
/*                                                                           */
/* This software is provided 'as-is', without any expressed or implied       */
/* warranty.  In no event will the authors be held liable for any damages    */
/* arising from the use of this software.                                    */
/*                                                                           */
/* Permission is granted to anyone to use this software for any purpose,     */
/* including commercial applications, and to alter it and redistribute it    */
/* freely, subject to the following restrictions:                            */
/*                                                                           */
/* 1. The origin of this software must not be misrepresented; you must not   */
/*    claim that you wrote the original software. If you use this software   */
/*    in a product, an acknowledgment in the product documentation would be  */
/*    appreciated but is not required.                                       */
/* 2. Altered source versions must be plainly marked as such, and must not   */
/*    be misrepresented as being the original software.                      */
/* 3. This notice may not be removed or altered from any source              */
/*    distribution.                                                          */
/*                                                                           */
/*****************************************************************************/



#ifndef COLOR_H
#define COLOR_H



/*****************************************************************************/
/*                                   Data                                    */
/*****************************************************************************/



/* Color structure */
typedef struct Color Color;
struct Color {
    unsigned char R;
    unsigned char G;
    unsigned char B;
    unsigned char A;
};



/*****************************************************************************/
/*                                   Code                                    */
/*****************************************************************************/


                                                             
#if defined(HAVE_INLINE)

INLINE Color RGB (unsigned char R, unsigned char G, unsigned char B)
/* Generate a color value */
{
    Color C;
    C.R = R; C.G = G; C.B = B; C.A = 0;
    return C;
}

INLINE Color RGBA (unsigned char R, unsigned char G, unsigned char B, unsigned char A)
/* Generate a color value */
{
    Color C;
    C.R = R; C.G = G; C.B = B; C.A = A;
    return C;
}

#else

Color RGB (unsigned char R, unsigned char G, unsigned char B);
/* Generate a color value */

Color RGBA (unsigned char R, unsigned char G, unsigned char B, unsigned char A);
/* Generate a color value */

#endif



/* End of color.h */

#endif
